/* $Id: daemonize.c,v 1.1 2002/12/05 11:10:41 loreti Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include "daemonize.h"

#define OPEN_MAX_GUESS    256

void daemonize(
  const char *ident
) {

/**
 | Calling this procedure, the current process is modified to run
 | as a daemon.
 |
 | See W. Richard Stevens, "Advanced Programming in the UNIX
 | Environment", chapter 13 (pages 415-425) for details.
 |
 | - Forks and terminates the original task;
 | - becomes the session (and group) leader of a new session
 |   (and process group), and removes the controlling terminal;
 | - clears the file mode creation mask;
 | - forks and terminates the original task;
 | - closes all file descriptors;
 | - connects stdin, stdout and stderr to /dev/null;
 | - and finally opens a connection to /var/log/messages.
**/

  static int openMax = OPEN_MAX;
  pid_t      pid;
  int        fd;

  if ((pid = fork()) < 0) {
    perror("fork 1 (daemonInit)");
    exit(EXIT_FAILURE);
  } else if (pid) {
    exit(EXIT_SUCCESS);         /* Parent; bye bye... */
  }

  setsid();
  chdir("/");
  umask(0);

/**
 | The second fork prevents acquisition of a controlling terminal
 | by guaranteeing that the process is not a session leader.  Under
 | BSD that happens only when an explicit ioctl request is made, but
 | under SVR4 it is automatic if a terminal device is opened by a
 | session leader and O_NOCTTY is not specified.
**/

  if ((pid = fork()) < 0) {
    perror("fork 2 (daemonInit)");
    exit(EXIT_FAILURE);
  } else if (pid) {             /* Parent */
    exit(EXIT_SUCCESS);
  }

/**
 | Kludge for OS's without OPEN_MAX (required by POSIX.1); see
 | W. Richard Stevens, "Advanced Programming in the UNIX Environment",
 | chapter 2, pages 42-44, for details.
**/

  if (!openMax) {
    errno = 0;
    if ((openMax = sysconf(_SC_OPEN_MAX)) < 0) {
      if (errno) {
        perror("sysconf");
        exit(EXIT_FAILURE);
      } else {
        openMax = OPEN_MAX_GUESS;
      }
    }
  }

/**
 | In a running daemon, all file descriptors have to be closed.
**/

  for (fd = 0;   fd < openMax;   ++fd) {
    close(fd);
  }

  if ((fd = open("/dev/null", O_RDWR)) < 0) {
    exit(EXIT_FAILURE);
  }

  if (fd != STDIN_FILENO            ||
      dup2(fd, STDOUT_FILENO) < 0   ||
      dup2(fd, STDERR_FILENO) < 0)
    {
      exit(EXIT_FAILURE);
    }

  openlog(ident, LOG_PID | LOG_CONS, LOG_DAEMON);
}
